#include<iostream>
#include<fstream>

using namespace std;

class Modell
{
    protected:
        int anzahl;
    public:
        Modell(int anzahl)
        {
            this->anzahl=anzahl;
        }


        virtual double rechteSeite(double* y, int zeile)=0;
        
        int holeAnzahlGleichungen()
        {
            return anzahl;
        } 
}; 


class ModellSchwinger : public Modell
{
    public:
        ModellSchwinger() : Modell(2)  
        {
        }
        double rechteSeite(double* y, int zeile)
        {
            switch(zeile)
            {
                case 0:
                     return y[1];
                break;
                case 1:
//                     return -y[0]-0.5*y[1]; 
                     return -y[0]; 
                break;                    
                default:
                     return 0.0;
                break;                                                                                            
            }               
        }
}; 

class ModellTilger : public Modell
{
    public:
        ModellTilger() : Modell(4)  
        {
            m1 = 3.0;
            C1 = 27.0;
            m2 = 1.0;
            C2 = 13.0;
            D2 = 4.0;                          
        }
        double rechteSeite(double* y, int zeile)
        {
            switch(zeile)
            {
                case 0:
                     return y[1];
                break;
                case 1:
                     return (-C1/m1-C2/m1)*y[0] + (-D2/m1)*y[1] + (C2/m1)*y[2] + (D2/m1)*y[3]; 
                break;                    
                case 2:
                     return y[3]; 
                break;                    
                case 3:
                     return (C2/m2)*y[0] + (D2/m2)*y[1] + (-C2/m2)*y[2] + (-D2/m2)*y[3]; 
                break;                    
                default:
                     return 0.0;
                break;                                                                                            
            }               
        }
        void setzeParameter(double C2, double D2)
        {
            this->C2 = C2;
            this->D2 = D2;             
        }
        
        double berechneGesamtenergie(double* y)
        {
            return 0.5*C1*y[0]*y[0]+0.5*C2*y[2]*y[2]+0.5*D2*y[3]*y[3];               
        }
    private:
        double m1,C1,m2,C2,D2;        
}; 


class Integrator
{
    public:
        Integrator(Modell* modell)
        {
            this->modell = modell;
            yneu = new double[modell->holeAnzahlGleichungen()];
        }
        virtual double* zeitschritt(double* yalt,double dt) = 0;
    protected:
        Modell* modell;
        double* yneu;
}; 


class IntegratorRuKu : public Integrator
{
    public:
        IntegratorRuKu(Modell* modell) : Integrator(modell)
        {
            yhilf = new double[modell->holeAnzahlGleichungen()];
            k1 = new double[modell->holeAnzahlGleichungen()]; 
            k2 = new double[modell->holeAnzahlGleichungen()]; 
            k3 = new double[modell->holeAnzahlGleichungen()]; 
            k4 = new double[modell->holeAnzahlGleichungen()]; 
        }
        double* zeitschritt(double* yalt,double dt)
        {
            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                k1[i] = modell->rechteSeite(yalt,i);
                yhilf[i] = yalt[i] + 0.5*dt*k1[i];
            }                
            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                k2[i] = modell->rechteSeite(yhilf,i);
                yhilf[i] = yalt[i] + 0.5*dt*k2[i];
            }                
            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                k3[i] = modell->rechteSeite(yhilf,i);
                yhilf[i] = yalt[i] + dt*k3[i];
            }                
            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                k4[i] = modell->rechteSeite(yhilf,i);
            }                

            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                yneu[i] = yalt[i] + (dt/6.0)*(k1[i]+2.0*k2[i]+2.0*k3[i]+k4[i]);
            }                

            return yneu;
        } 
    private:
        double* yhilf;
        double *k1,*k2,*k3,*k4;
}; 

class IntegratorEuler : public Integrator
{
    public:
        IntegratorEuler(Modell* modell) : Integrator(modell)
        {
        }
        double* zeitschritt(double* yalt,double dt)
        {
            for(int i=0;i<modell->holeAnzahlGleichungen();i++)
            {
                f = modell->rechteSeite(yalt,i);
                yneu[i] = yalt[i] + f*dt;
            }
            return yneu;
        }
    private:
        double f;                     
}; 

class Fehler
{
      public:
          Fehler()
          {
              modell          = new ModellTilger();
              integrator  = new IntegratorRuKu(modell);
          }
          
          double berechneFehler(double* parameter)
          {
              modell->setzeParameter(parameter[0],parameter[1]);

              double t=0.0;
              double dt=0.005;

              double* yneu;
              double yalt[4] = {0.5,0.0,0.0,0.0};

              for(int i=0;i<2001;i++)
              {
                  yneu = integrator->zeitschritt(yalt,dt);

                  yalt[0] = yneu[0];
                  yalt[1] = yneu[1];
                  yalt[2] = yneu[2];
                  yalt[3] = yneu[3];

                  t=t+dt;
              }
              
              double erg = modell->berechneGesamtenergie(yneu);
              if(erg<0.0)
                  return 99999999.0;
              else
                  return erg;                  
          }
      private:
          ModellTilger*     modell;
          IntegratorRuKu*   integrator;                        
};

class Optimierer
{
    public:
        Optimierer(Fehler* fehler,int parameteranzahl, double* startparameter)
        {
            this->fehler =  fehler;
            panz   =  parameteranzahl;
            param      =  new double[parameteranzahl];
            schrittweite = new double[parameteranzahl];
            fehler_neu = new double*[parameteranzahl];
            for(int i=0;i<parameteranzahl;i++)
            {
                param[i] = startparameter[i];
                schrittweite[i] = 1.0;
                fehler_neu[i] = new double[7];
            }                
        }
        
        double* starten(int itermax)
        {
            for(int i=0;i<itermax;i++)
            {
                wert = fehler->berechneFehler(param);                        
                cout<<"Iteration Nr."<<i<<" Fehler="<<wert<<" C2="<<param[0]<<" D2="<<param[1]<<endl;

                for(int k=0;k<panz;k++)
                {
                    param[k]-=2.0*schrittweite[k];
                    fehler_neu[k][0] = fehler->berechneFehler(param);                        
                    param[k]+=2.0*schrittweite[k];
                    
                    param[k]-=1.0*schrittweite[k];
                    fehler_neu[k][1] = fehler->berechneFehler(param);                        
                    param[k]+=1.0*schrittweite[k];
                    
                    param[k]-=0.5*schrittweite[k];
                    fehler_neu[k][2] = fehler->berechneFehler(param);                        
                    param[k]+=0.5*schrittweite[k];

                    fehler_neu[k][3] = fehler->berechneFehler(param);                        
                    
                    param[k]+=0.5*schrittweite[k];
                    fehler_neu[k][4] = fehler->berechneFehler(param);                        
                    param[k]-=0.5*schrittweite[k];

                    param[k]+=1.0*schrittweite[k];
                    fehler_neu[k][5] = fehler->berechneFehler(param);                        
                    param[k]-=1.0*schrittweite[k];

                    param[k]+=2.0*schrittweite[k];
                    fehler_neu[k][6] = fehler->berechneFehler(param);                        
                    param[k]-=2.0*schrittweite[k];
                }
                
                for(int k=0;k<panz;k++)
                {
                    if( istKleinster(0,fehler_neu[k]) )
                    {
                         param[k]-=2.0*schrittweite[k];                        
                         schrittweite[k]*=2.0;
                    }                        
                    else if( istKleinster(1,fehler_neu[k]) )
                    {
                         param[k]-=1.0*schrittweite[k];                        
                         schrittweite[k]*=1.0;
                    }
                    else if( istKleinster(2,fehler_neu[k]) )
                    {
                         param[k]-=0.5*schrittweite[k];                        
                         schrittweite[k]*=0.5;
                    }
                    else if( istKleinster(3,fehler_neu[k]) )
                    {
                         param[k]=param[k];                        
                    }
                    else if( istKleinster(4,fehler_neu[k]) )
                    {
                         param[k]+=0.5*schrittweite[k];                        
                         schrittweite[k]*=0.5;
                    }
                    else if( istKleinster(5,fehler_neu[k]) )
                    {
                         param[k]+=1.0*schrittweite[k];                        
                         schrittweite[k]*=1.0;
                    }
                    else if( istKleinster(6,fehler_neu[k]) )
                    {
                         param[k]+=2.0*schrittweite[k];                        
                         schrittweite[k]*=2.0;
                    }
                }
            }
            return param;
        }              
    private:
        bool istKleinster(int index,double* ff)
        {
            double wert = ff[index];
            for(int i=0;i<7;i++)
                if(index!=i && ff[i]<wert)
                    return false;             
        }            
            
            
        Fehler* fehler;
        int panz;
        double* param;
        double* schrittweite;                    
        double** fehler_neu;
        double wert;
};


int main()
{
    double p0[2] = {13.0,4.0};
    Fehler*     fehler          = new Fehler();
    Optimierer*  optimierer      = new Optimierer(fehler,2,p0);
    
    double* pbest = optimierer->starten(100);
    
    cout<<"Beste Kombination C2, D2 nach 100 Optimierungsschritten:"<<endl;
    cout<<"C2="<<pbest[0]<<" D2="<<pbest[1]<<endl;
    
}
